@@ -13,4 +13,5 @@ together.
 
 from __future__ import absolute_import
 
-from .core import OdeSys
+from ._release import __version__
+from .core import ODESys, OdeSys
