@@ -7,13 +7,21 @@ from math import log
 import numpy as np
 
 
-def plot_result(x, y, indices=None, plot=None, plot_kwargs_cb=None, ax=None,
+def _set_scale(cb, argstr):
+    if argstr.count(';') == 0:
+        cb(argstr)
+    else:
+        arg, kw = argstr.split(';')
+        cb(arg, **eval('dict(%s)' % kw))
+
+
+def plot_result(x, y, indices=None, plot_kwargs_cb=None, ax=None,
                 ls=('-', '--', ':', '-.'),
                 c=('k', 'r', 'g', 'b', 'c', 'm', 'y'),
                 m=('o', 'v', '8', 's', 'p', 'x', '+', 'd', 's'),
                 m_lim=-1, lines=None, interpolate=None, interp_from_deriv=None,
                 names=None, latex_names=None, xlabel=None, ylabel=None,
-                xscale=None, yscale=None):
+                xscale=None, yscale=None, legend=False, yerr=None):
     """
     Plot the depepndent variables vs. the independent variable
 
@@ -56,11 +64,8 @@ def plot_result(x, y, indices=None, plot=None, plot_kwargs_cb=None, ax=None,
     """
     import matplotlib.pyplot as plt
 
-    if plot is None:
-        if ax is None:
-            from matplotlib.pyplot import plot
-        else:
-            plot = ax.plot
+    if ax is None:
+        ax = plt.subplot(1, 1, 1)
     if plot_kwargs_cb is None:
         def plot_kwargs_cb(idx, lines=False, markers=False, labels=None):
 
@@ -95,11 +100,23 @@ def plot_result(x, y, indices=None, plot=None, plot_kwargs_cb=None, ax=None,
     if lines is None:
         lines = interpolate in (None, False)
     markers = len(x) < m_lim
+
+    if yerr is not None:
+        for idx in indices:
+            clr = plot_kwargs_cb(idx)['c']
+            ax.fill_between(x, _y[:, idx] - yerr[:, idx], _y[:, idx] + yerr[:, idx], facecolor=clr, alpha=.3)
+
+    if isinstance(yscale, str) and 'linthreshy' in yscale:
+        arg, kw = yscale.split(';')
+        thresh = eval('dict(%s)' % kw)['linthreshy']
+        ax.axhline(thresh, linewidth=.5, linestyle='--', color='k', alpha=.5)
+        ax.axhline(-thresh, linewidth=.5, linestyle='--', color='k', alpha=.5)
+
     for idx in indices:
-        plot(x, _y[:, idx], **plot_kwargs_cb(
+        ax.plot(x, _y[:, idx], **plot_kwargs_cb(
             idx, lines=lines, labels=latex_names or names))
         if markers:
-            plot(x, _y[:, idx], **plot_kwargs_cb(
+            ax.plot(x, _y[:, idx], **plot_kwargs_cb(
                 idx, lines=False, markers=markers, labels=latex_names or names))
 
     if xlabel is None:
@@ -145,14 +162,22 @@ def plot_result(x, y, indices=None, plot=None, plot_kwargs_cb=None, ax=None,
             y2[:, idx] = interp_cb(x_plot)
 
         for idx in indices:
-            plot(x_plot, y2[:, idx], **plot_kwargs_cb(
+            ax.plot(x_plot, y2[:, idx], **plot_kwargs_cb(
                 idx, lines=True, markers=False))
         return x_plot, y2
 
     if xscale is not None:
-        (ax or plt.gca()).set_xscale(xscale)
+        _set_scale(ax.set_xscale, xscale)
     if yscale is not None:
-        (ax or plt.gca()).set_yscale(yscale)
+        _set_scale(ax.set_yscale, yscale)
+
+    if legend is True:
+        ax.legend()
+    elif legend in (None, False):
+        pass
+    else:
+        ax.legend(**legend)
+    return ax
 
 
 def plot_phase_plane(x, y, indices=None, plot=None, names=None, **kwargs):
@@ -237,6 +262,7 @@ def info_vlines(ax, xout, info, vline_colors=('maroon', 'purple'),
             ln_np1 = log(len(vlines)+1)
             every = min(round((ln_np1 - 4)/log(2)), 1)
 
-        ax.vlines(vlines[::every], idx/nvk + 0.002, (idx+1)/nvk - 0.002, colors=vline_colors[idx % len(vline_colors)],
+        ax.vlines(vlines[::every], idx/nvk + 0.002, (idx+1)/nvk - 0.002,
+                  colors=vline_colors[idx % len(vline_colors)],
                   alpha=alpha, transform=ax.get_xaxis_transform())
     right_hand_ylabels(ax, [k[3] if k.startswith('fe_') else k[0] for k in vline_keys])
