@@ -10,11 +10,6 @@ import numpy as np
 import pytest
 import time
 
-try:
-    import sympy as sp
-except ImportError:
-    sp = None
-
 try:
     import sym
 except ImportError:
@@ -25,7 +20,7 @@ else:
 
 from .. import ODESys
 from ..core import integrate_chained
-from ..symbolic import SymbolicSys, ScaledSys, symmetricsys, PartiallySolvedSystem, get_logexp
+from ..symbolic import SymbolicSys, ScaledSys, symmetricsys, PartiallySolvedSystem, get_logexp, _group_invariants
 from ..util import requires
 from .bateman import bateman_full  # analytic, never mind the details
 from .test_core import vdp_f
@@ -38,6 +33,31 @@ def identity(x):
 idty2 = (identity, identity)
 
 
+def _decay3(x, y, p):
+    return [
+        -p[0]*y[0],
+        p[0]*y[0] - p[1]*y[1],
+        p[1]*y[1] - p[2]*y[2]
+    ]
+
+
+def _get_decay3(**kwargs):
+    return SymbolicSys.from_callback(_decay3, 3, 3, **kwargs)
+
+
+def _get_decay3_names(yn, pn, **kwargs):
+    def f(x, y, p):
+        y = [y[n] for n in yn]
+        p = [p[n] for n in pn]
+        return dict(zip(yn, [
+            -p[0]*y[0],
+            p[0]*y[0] - p[1]*y[1],
+            p[1]*y[1] - p[2]*y[2]
+        ]))
+    return SymbolicSys.from_callback(f, names=yn, param_names=pn, dep_by_name=True,
+                                     par_by_name=True, indep_name='t', **kwargs)
+
+
 @requires('sym', 'scipy')
 def test_SymbolicSys():
     from pyodesys.integrators import RK4_example_integrator
@@ -63,6 +83,28 @@ def test_SymbolicSys():
         SymbolicSys.from_callback(lambda x, y, p, be: [], 2, names=['foo', 'bar'])
 
 
+@requires('sym', 'scipy')
+def test_SymbolicSys__indep_name():
+    odesys = SymbolicSys.from_callback(
+        lambda t, y, p: {
+            'x': -p['a']*y['x'],
+            'y': -p['b']*y['y'] + p['a']*y['x'],
+            'z': p['b']*y['y']
+        }, names='xyz', param_names='ab', dep_by_name=True, par_by_name=True)
+    pars = {'a': [11, 17, 19], 'b': 13}
+    results = odesys.integrate([42, 43, 44], {'x': 7, 'y': 5, 'z': 3}, pars)
+    for r, a in zip(results, pars['a']):
+        assert np.allclose(r.named_dep('x'), 7*np.exp(-a*(r.xout - r.xout[0])))
+
+
+@requires('sym')
+def test_SymbolicSys__init_indep__init_dep():
+    odesys = SymbolicSys.from_callback(lambda x, y, p, be: [-y[0], y[0]], 2, names=['foo', 'bar'], indep_name='t',
+                                       init_indep=True, init_dep=True)
+    assert odesys.init_indep.name == 'i_t'
+    assert [dep.name for dep in odesys.init_dep] == ['i_foo', 'i_bar']
+
+
 def decay_rhs(t, y, k):
     ny = len(y)
     dydt = [0]*ny
@@ -106,6 +148,7 @@ def test_TransformedSys_logy_logx():
 
 @requires('sym', 'pycvodes', 'sympy')
 def test_TransformedSys_logy_logx_scaled_shifted():
+    import sympy as sp
     em16 = (sp.S.One*10)**-16
     _test_TransformedSys(get_logexp(42, em16), get_logexp(42, em16), 1e-7, 1e-7, 1e-4,
                          150, y_zero=0, t_zero=0, nsteps=800)
@@ -123,6 +166,7 @@ def test_TransformedSys_liny_logx():
 
 @requires('sym', 'pycvodes')
 def test_ScaledSys():
+    import sympy as sp
     k = k0, k1, k2 = [7., 3, 2]
     y0, y1, y2, y3 = sp.symbols('y0 y1 y2 y3', real=True, positive=True)
     # this is actually a silly example since it is linear
@@ -214,6 +258,7 @@ def timeit(callback, *args, **kwargs):
 @requires('sym', 'pyodeint')
 @pytest.mark.parametrize('method', ['bs', 'rosenbrock4'])
 def test_exp(method):
+    import sympy as sp
     x = sp.Symbol('x')
     symsys = SymbolicSys([(x, sp.exp(x))])
     tout = [0, 1e-9, 1e-7, 1e-5, 1e-3, 0.1]
@@ -226,6 +271,7 @@ def test_exp(method):
 
 # @pytest.mark.xfail
 def _test_mpmath():  # too slow
+    import sympy as sp
     x = sp.Symbol('x')
     symsys = SymbolicSys([(x, sp.exp(x))])
     tout = [0, 1e-9, 1e-7, 1e-5, 1e-3, 0.1]
@@ -281,6 +327,7 @@ def test_SymbolicSys__from_callback_bateman(band):
 
 
 def _test_bateman(SymbSys, **kwargs):
+    import sympy as sp
     tend, k, y0 = 2, [4, 3], (5, 4, 2)
     y = sp.symarray('y', len(k)+1)
     dydt = decay_dydt_factory(k)
@@ -372,6 +419,7 @@ def test_scipy(name, forgive):
 
 
 # (dopri5, .2), (bs, .03) <-- works in boost 1.59
+@pytest.mark.slow
 @requires('sym', 'pyodeint')
 @pytest.mark.parametrize('method,forgive', zip(
     'rosenbrock4 dopri5 bs'.split(), (.2, .2, .04)))
@@ -400,6 +448,7 @@ def _gsl(tout, method, forgive):
     check(yout[-1, :], n, p, a, atol, rtol, forgive)
 
 
+@pytest.mark.veryslow
 @requires('sym', 'pygslodeiv2')
 @pytest.mark.parametrize('method,forgive', zip(
     'msadams msbdf rkck bsimp'.split(), (5, 14, 0.2, 0.02)))
@@ -407,6 +456,7 @@ def test_gsl_predefined(method, forgive):
     _gsl([10**i for i in range(-14, 1)], method, forgive)
 
 
+@pytest.mark.veryslow
 @requires('sym', 'pygslodeiv2')
 @pytest.mark.parametrize('method,forgive', zip(
     'bsimp msadams msbdf rkck'.split(), (0.01, 4, 14, 0.21)))
@@ -426,6 +476,7 @@ def _cvode(tout, method, forgive):
     check(yout[-1, :], n, p, a, atol, rtol, forgive)
 
 
+@pytest.mark.slow
 @requires('sym', 'pycvodes')
 @pytest.mark.parametrize('method,forgive', zip(
     'adams bdf'.split(), (2.4, 5.0)))
@@ -434,6 +485,7 @@ def test_cvode_predefined(method, forgive):
 
 
 # cvode performs significantly better than vode:
+@pytest.mark.slow
 @requires('sym', 'pycvodes')
 @pytest.mark.parametrize('method,forgive', zip(
     'adams bdf'.split(), (2.4, 5)))
@@ -441,6 +493,7 @@ def test_cvode_adaptive(method, forgive):
     _cvode(1, method, forgive)
 
 
+@pytest.mark.veryslow
 @requires('sym', 'scipy')
 @pytest.mark.parametrize('n,forgive', [(4, 1), (17, 1), (42, 7)])
 def test_long_chain_dense(n, forgive):
@@ -453,6 +506,7 @@ def test_long_chain_dense(n, forgive):
     check(yout[-1, :], n, p, a, atol, rtol, forgive)
 
 
+@pytest.mark.slow
 @requires('sym', 'scipy')
 @pytest.mark.parametrize('n', [29])  # something maxes out at 31
 def test_long_chain_banded_scipy(n):
@@ -484,6 +538,7 @@ def test_long_chain_banded_scipy(n):
     assert min_time_dens*2 > min_time_band  # (2x: fails sometimes due to load)
 
 
+@pytest.mark.veryslow
 @requires('sym', 'pycvodes')
 @pytest.mark.parametrize('n', [29, 79])
 def test_long_chain_banded_cvode(n):
@@ -513,15 +568,7 @@ def test_long_chain_banded_cvode(n):
         pass  # will fail sometimes due to load
 
 
-def _get_decay3(**kwargs):
-    return SymbolicSys.from_callback(
-        lambda x, y, p: [
-            -p[0]*y[0],
-            p[0]*y[0] - p[1]*y[1],
-            p[1]*y[1] - p[2]*y[2]
-        ], 3, 3, **kwargs)
-
-
+@pytest.mark.slow
 @requires('sym', 'pycvodes', 'pygslodeiv2')
 @pytest.mark.parametrize('integrator', ['cvode', 'gsl'])
 def test_no_diff_adaptive_chained_single(integrator):
@@ -540,6 +587,7 @@ def test_no_diff_adaptive_chained_single(integrator):
     assert np.allclose(yout2, ref)
 
 
+@pytest.mark.slow
 @requires('sym', 'pycvodes', 'pygslodeiv2')
 @pytest.mark.parametrize('integrator', ['cvode', 'gsl'])
 def test_no_diff_adaptive_chained_single__multimode(integrator):
@@ -550,9 +598,8 @@ def test_no_diff_adaptive_chained_single__multimode(integrator):
     _k = [3.5, 2.5, 1.5]
     k = [_k]*4
     res1 = odesys.integrate(tout, y0, k, integrator=integrator, first_step=1e-14)
-    for xout1, yout1, info1 in zip(*res1):
-        assert np.allclose(xout1 - xout1[0], res1[0][0] - res1[0][0][0])
-        assert np.allclose(yout1, res1[1][0])
+    for res in res1:
+        xout1, yout1, info1 = res.xout, res.yout, res.info
         ref = np.array(bateman_full(_y0, _k, xout1 - xout1[0], exp=np.exp)).T
         assert info1['success']
         assert xout1.size > 10
@@ -560,8 +607,9 @@ def test_no_diff_adaptive_chained_single__multimode(integrator):
         assert np.allclose(yout1, ref)
 
     res2 = integrate_chained([odesys], {}, tout, y0, k, integrator=integrator, first_step=1e-14)
-    for xout2, yout2, info2 in zip(*res2):
-        assert info1['success']
+    for res in res2:
+        xout2, yout2, info2 = res.xout, res.yout, res.info
+        assert info2['success']
         assert xout2.size == xout1.size
         assert np.allclose(yout2, ref)
 
@@ -575,12 +623,57 @@ def test_PartiallySolvedSystem(integrator):
     })
     y0 = [3, 2, 1]
     k = [3.5, 2.5, 1.5]
-    xout, yout, info = partsys.integrate([0, 1], y0, k, integrator=integrator)
+    xout, yout, info = partsys.integrate(
+        [0, 1], y0, k, integrator=integrator)
     ref = np.array(bateman_full(y0, k, xout - xout[0], exp=np.exp)).T
     assert info['success']
     assert np.allclose(yout, ref)
 
 
+@requires('sym', 'pycvodes')
+def test_PartiallySolvedSystem_ScaledSys():
+    odesys = _get_decay3(lower_bounds=[0, 0, 0])
+    partsys = PartiallySolvedSystem(odesys, lambda x0, y0, p0, be: {
+        odesys.dep[0]: y0[0]*be.exp(-p0[0]*(odesys.indep-x0))
+    })
+    y0 = [3, 2, 1]
+    k = [3.5, 2.5, 1.5]
+
+    def _check(res):
+        ref = np.array(bateman_full(y0, k, res.xout - res.xout[0], exp=np.exp)).T
+        assert res.info['success']
+        assert np.allclose(res.yout, ref)
+    args = [0, 1], y0, k
+    kwargs = dict(integrator='cvode')
+    _check(odesys.integrate(*args, **kwargs))
+    _check(partsys.integrate(*args, **kwargs))
+    scaledsys = ScaledSys.from_other(partsys, dep_scaling=42, indep_scaling=17)
+    _check(scaledsys.integrate(*args, **kwargs))
+
+
+@requires('sym', 'pycvodes', 'pygslodeiv2')
+@pytest.mark.parametrize('integrator', ['cvode', 'gsl'])
+def test_PartiallySolvedSystem_multi(integrator):
+    odesys = _get_decay3()
+
+    def _get_analytic(x0, y0, p0, be):
+        a0 = y0[0]*be.exp(-p0[0]*(odesys.indep - x0))
+        a1 = y0[0] + y0[1] + y0[2] - a0 - odesys.dep[2]
+        return [a0, a1]
+
+    def subst(x0, y0, p0, be):
+        a0, a1 = _get_analytic(x0, y0, p0, be)
+        return {odesys.dep[0]: a0, odesys.dep[1]: a1}
+
+    partsys = PartiallySolvedSystem(odesys, subst)
+    a0, a1 = _get_analytic(partsys.init_indep,
+                           partsys.init_dep,
+                           odesys.params,
+                           odesys.be)
+    assert partsys.ny == 1
+    assert partsys.exprs[0].subs(odesys.params[2], 0) - odesys.params[1]*a1 == 0
+
+
 @requires('sym', 'pycvodes', 'pygslodeiv2')
 @pytest.mark.parametrize('integrator', ['cvode', 'gsl'])
 def test_PartiallySolvedSystem__using_y(integrator):
@@ -597,6 +690,7 @@ def test_PartiallySolvedSystem__using_y(integrator):
     assert np.allclose(np.sum(yout, axis=1), sum(y0))
 
 
+@pytest.mark.slow
 @requires('sym', 'pycvodes', 'pygslodeiv2')
 @pytest.mark.parametrize('integrator', ['cvode', 'gsl'])
 def test_PartiallySolvedSystem_multiple_subs(integrator):
@@ -632,15 +726,19 @@ def test_PartiallySolvedSystem_multiple_subs__transformed(integrator):
     y0 = [3, 2, 1]
     k = [3.5, 2.5, 0]
     tend = 1
-    xout, yout, info = loglogpartsys.integrate([1e-12, tend], y0, k, integrator=integrator, first_step=1e-14)
-    ref = np.array(bateman_full(y0, k, xout - xout[0], exp=np.exp)).T
-    assert info['success']
-    assert info['internal_yout'].shape[-1] == 1
-    assert info['internal_yout'][-1, 0] < 0  # ln(y[1])
-    assert np.allclose(yout, ref)
+    for system, ny_internal in [(odesys, 3), (partsys, 1), (loglogpartsys, 1)]:
+        xout, yout, info = system.integrate([1e-12, tend], y0, k, integrator=integrator,
+                                            first_step=1e-14, nsteps=1000)
+        ref = np.array(bateman_full(y0, k, xout - xout[0], exp=np.exp)).T
+        assert info['success']
+        assert info['internal_yout'].shape[-1] == ny_internal
+        if system == loglogpartsys:
+            assert info['internal_yout'][-1, 0] < 0  # ln(y[1])
+        assert np.allclose(yout, ref)
 
 
 def _get_transf_part_system():
+    import sympy as sp
     odesys = _get_decay3()
     partsys = PartiallySolvedSystem(odesys, lambda x0, y0, p0: {
         odesys.dep[0]: y0[0]*sp.exp(-p0[0]*(odesys.indep-x0))
@@ -649,6 +747,7 @@ def _get_transf_part_system():
     return LogLogSys.from_other(partsys)
 
 
+@pytest.mark.slow
 @requires('sym', 'pycvodes', 'pygslodeiv2')
 @pytest.mark.parametrize('integrator', ['cvode', 'gsl'])
 def test_PartiallySolvedSystem__symmetricsys(integrator):
@@ -662,18 +761,20 @@ def test_PartiallySolvedSystem__symmetricsys(integrator):
     assert np.allclose(np.sum(yout, axis=1), sum(y0))
 
 
+@pytest.mark.slow
 @requires('sym', 'pycvodes', 'pygslodeiv2')
 @pytest.mark.parametrize('integrator', ['cvode', 'gsl'])
 def test_PartiallySolvedSystem__symmetricsys__multi(integrator):
     trnsfsys = _get_transf_part_system()
     y0s = [[3., 2., 1.], [3.1, 2.1, 1.1], [3.2, 2.3, 1.2], [3.6, 2.4, 1.3]]
     ks = [[3.5, 2.5, 0], [3.3, 2.4, 0], [3.2, 2.1, 0], [3.3, 2.4, 0]]
-    xout, yout, info = trnsfsys.integrate([(1e-10, 1)]*len(ks), y0s, ks, integrator=integrator, first_step=1e-14)
+    results = trnsfsys.integrate([(1e-10, 1)]*len(ks), y0s, ks, integrator=integrator, first_step=1e-14)
     for i, (y0, k) in enumerate(zip(y0s, ks)):
-        ref = np.array(bateman_full(y0, k, xout[i] - xout[i][0], exp=np.exp)).T
-        assert info[i]['success'] and info[i]['nfev'] > 10
-        assert info[i]['nfev'] > 1 and info[i]['time_cpu'] < 100
-        assert np.allclose(yout[i], ref) and np.allclose(np.sum(yout[i], axis=1), sum(y0))
+        xout, yout, info = results[i]
+        ref = np.array(bateman_full(y0, k, xout - xout[0], exp=np.exp)).T
+        assert info['success'] and info['nfev'] > 10
+        assert info['nfev'] > 1 and info['time_cpu'] < 100
+        assert np.allclose(yout, ref) and np.allclose(np.sum(yout, axis=1), sum(y0))
 
 
 def _get_nonlin(**kwargs):
@@ -707,6 +808,7 @@ def _ref_nonlin(y0, k, t):
     return (x4*x8 + x5*x8 + x7*(x2 + x6))/(2*kf*(x6 + x7))
 
 
+@pytest.mark.slow
 @requires('sym', 'pycvodes', 'pygslodeiv2')
 @pytest.mark.parametrize('integrator', ['cvode', 'gsl'])
 def test_PartiallySolvedSystem__symmetricsys__nonlinear(integrator):
@@ -799,6 +901,7 @@ def test_SymbolicSys_from_callback__symcxx():
     _test_SymbolicSys_from_callback__backend('symcxx')
 
 
+@pytest.mark.slow
 @requires('sym', 'pycvodes', 'pygslodeiv2')
 @pytest.mark.parametrize('integrator,method', [('cvode', 'adams'), ('gsl', 'msadams')])
 def test_integrate_chained(integrator, method):
@@ -817,7 +920,7 @@ def test_integrate_chained(integrator, method):
 
         xout, yout, info = integrate_chained([logsys, linsys], {'nsteps': [1, 1]}, tout, y0,
                                              return_on_error=True, **kw)
-        assert info['success'] is False
+        assert info['success'] == False  # noqa
         ntot = 400
         nlinear = 60*(p+3)
 
@@ -826,7 +929,7 @@ def test_integrate_chained(integrator, method):
             'first_step': [30.0, 1e-5],
             'return_on_error': [True, False]
         }, tout, y0, **kw)
-        assert info['success'] is True
+        assert info['success'] == True  # noqa
         check(yout[-1, :], n, p, a, atol, rtol, forgive)
 
 
@@ -844,13 +947,13 @@ def _test_cetsa(y0, params, extra=False, stepx=1, **kwargs):
 
     comb_res = integrate_chained([tsys, odesys], {'nsteps': [500*stepx, 20*stepx]}, tout, y0/molar_unitless, params,
                                  return_on_error=True, autorestart=2, **kwargs)
-    if isinstance(comb_res[2], dict):
-        assert comb_res[2]['success']
-        assert comb_res[2]['nfev'] > 10
+    if isinstance(comb_res, list):
+        for r in comb_res:
+            assert r.info['success']
+            assert r.info['nfev'] > 10
     else:
-        for d in comb_res[2]:
-            assert d['success']
-            assert d['nfev'] > 10
+        assert comb_res.info['success']
+        assert comb_res.info['nfev'] > 10
 
     if extra:
         with pytest.raises(RuntimeError):  # (failure)
@@ -859,12 +962,13 @@ def _test_cetsa(y0, params, extra=False, stepx=1, **kwargs):
 
         res = odesys.integrate(np.linspace(t0, tend, 20), y0/molar_unitless, params, nsteps=int(38*1.1),
                                first_step=1e-14, **kwargs)
-        assert np.min(res[1][-1, :]) < -1e-6  # crazy! (failure of the linear formulation)
+        assert np.min(res.yout[-1, :]) < -1e-6  # crazy! (failure of the linear formulation)
         tres = tsys.integrate([t0, tend], y0/molar_unitless, params, nsteps=int(1345*1.1), **kwargs)
-        assert tres[2]['success'] is True
-        assert tres[2]['nfev'] > 100
+        assert tres.info['success'] is True
+        assert tres.info['nfev'] > 100
 
 
+@pytest.mark.veryslow
 @requires('sym', 'pycvodes', 'pygslodeiv2', 'pyodeint')
 @pytest.mark.parametrize('integrator', ['cvode', 'gsl', 'odeint'])
 def test_cetsa(integrator):
@@ -874,6 +978,7 @@ def test_cetsa(integrator):
         _test_cetsa(_cetsa.ys[0], _cetsa.ps[0], extra=True, integrator=integrator)
 
 
+@pytest.mark.veryslow
 @requires('sym', 'pycvodes', 'pygslodeiv2', 'pyodeint')
 @pytest.mark.parametrize('integrator', ['cvode', 'gsl', 'odeint'])
 def test_cetsa_multi(integrator):
@@ -954,6 +1059,7 @@ def test_SymbolicSys__first_step_expr():
     assert np.allclose(yout, ref, atol=10*kwargs['atol'], rtol=10*kwargs['rtol'])
 
 
+@pytest.mark.slow
 @requires('sym', 'pygslodeiv2')
 def test_SymbolicSys__from_callback__first_step_expr():
     tend, k, y0 = 5, [1e23, 3], (.7, .0, .0)
@@ -980,9 +1086,13 @@ def test_SymbolicSys__from_callback__first_step_expr__by_name():
         first_step_factory=lambda x0, ic: 1e-30*ic['foo'])
     y0 = {'foo': .7, 'bar': 0, 'baz': 0}
     p = {'first': 1e23, 'second': 2, 'third': 3}
-    xout, yout, info = odesys.integrate(5, y0, p, **kwargs)
-    ref = np.array(bateman_full([y0[k] for k in names], [p[k] for k in par_names], xout - xout[0], exp=np.exp)).T
-    assert np.allclose(yout, ref, atol=10*kwargs['atol'], rtol=10*kwargs['rtol'])
+    result = odesys.integrate(5, y0, p, **kwargs)
+    assert result.info['success']
+    ref = bateman_full([y0[k] for k in names], [p[k] for k in par_names], result.xout - result.xout[0], exp=np.exp)
+    for i, k in enumerate(odesys.names):
+        assert np.allclose(result.named_dep(k), ref[i], atol=10*kwargs['atol'], rtol=10*kwargs['rtol'])
+    for k, v in p.items():
+        assert result.named_param(k) == v
 
 
 @requires('sym', 'pyodeint')
@@ -1016,6 +1126,55 @@ def test_PartiallySolvedSystem__by_name():
         assert yout.ndim == 2 and xout.ndim == 1
 
 
+@requires('sym', 'pycvodes')
+def test_PartiallySolvedSystem__by_name_2():
+    yn, pn = 'x y z'.split(), 'p q r'.split()
+    odesys = _get_decay3_names(yn, pn)
+    partsys = PartiallySolvedSystem(odesys, lambda x0, y0, p0, be: {
+        odesys['x']: y0[odesys['x']]*be.exp(-p0['p']*(odesys.indep-x0))
+    })
+    y0 = [3, 2, 1]
+    k = [3.5, 2.5, 1.5]
+
+    def _check(res):
+        ref = np.array(bateman_full(y0, k, res.xout - res.xout[0], exp=np.exp)).T
+        assert res.info['success']
+        assert np.allclose(res.yout, ref)
+    args = [0, 1], dict(zip(yn, y0)), dict(zip(pn, k))
+    kwargs = dict(integrator='cvode')
+    _check(odesys.integrate(*args, **kwargs))
+    _check(partsys.integrate(*args, **kwargs))
+    scaledsys = ScaledSys.from_other(partsys, dep_scaling=42, indep_scaling=17)
+    _check(scaledsys.integrate(*args, **kwargs))
+
+
+@requires('sym')
+def test_symmetricsys__invariants():
+    yn, pn = 'x y z'.split(), 'a b'.split()
+    odesys = SymbolicSys.from_callback(
+        lambda t, y, p: {
+            'x': -p['a']*y['x'],
+            'y': -p['b']*y['y'] + p['a']*y['x'],
+            'z': p['b']*y['y']
+        }, names=yn, param_names=pn, dep_by_name=True, par_by_name=True,
+        linear_invariants=[[1, 1, 1]], linear_invariant_names=['mass-conservation'],
+        indep_name='t')
+    assert odesys.linear_invariants.tolist() == [[1, 1, 1]]
+    assert odesys.linear_invariant_names == ['mass-conservation']
+    assert odesys.nonlinear_invariants is None
+    assert odesys.nonlinear_invariant_names is None
+
+    logexp = get_logexp()
+    LogLogSys = symmetricsys(logexp, logexp)
+    tsys = LogLogSys.from_other(odesys)
+    assert tsys.linear_invariants is None
+    assert tsys.linear_invariant_names is None
+    assert len(tsys.nonlinear_invariants) == 1
+    E = odesys.be.exp
+    assert tsys.nonlinear_invariants[0] - sum(E(odesys[k]) for k in yn) == 0
+    assert tsys.nonlinear_invariant_names == ['mass-conservation']
+
+
 @requires('sym', 'pycvodes')
 def test_SymbolicSys__roots():
     def f(t, y):
@@ -1035,7 +1194,7 @@ def test_SymbolicSys__roots():
 def test_SymbolicSys__reference_parameters_using_symbols(method):
     be = sym.Backend('sympy')
     x, p = map(be.Symbol, 'x p'.split())
-    symsys = SymbolicSys([(x, -p*x)])
+    symsys = SymbolicSys([(x, -p*x)], params=True)
     tout = [0, 1e-9, 1e-7, 1e-5, 1e-3, 0.1]
     for y_symb in [False, True]:
         for p_symb in [False, True]:
@@ -1054,8 +1213,8 @@ def test_SymbolicSys__reference_parameters_using_symbols_from_callback(method):
     def dydt(t, y):       # external symbolic parameter 'k', should be allowed
         return [-k*y[0]]  # even though reminiscent of global variables.
 
-    odesys1 = SymbolicSys.from_callback(dydt, 1, backend=be)
-    odesys2 = SymbolicSys.from_callback(dydt, 1, backend=be, par_by_name=True, param_names=[])
+    odesys1 = SymbolicSys.from_callback(dydt, 1, backend=be, params=True)
+    odesys2 = SymbolicSys.from_callback(dydt, 1, backend=be, par_by_name=True, param_names=[], params=True)
     tout = [0, 1e-9, 1e-7, 1e-5, 1e-3, 0.1]
     for symsys in (odesys1, odesys2):
         for y_symb in [False, True]:
@@ -1067,6 +1226,7 @@ def test_SymbolicSys__reference_parameters_using_symbols_from_callback(method):
                 assert np.allclose(yout[:, 0], 2*np.exp(-3*xout))
 
 
+@pytest.mark.slow
 @requires('sym', 'pycvodes')
 @pytest.mark.parametrize('scaled', [False, True])
 def test_PartiallySolvedSystem__from_linear_invariants(scaled):
@@ -1095,6 +1255,11 @@ def test_PartiallySolvedSystem__from_linear_invariants(scaled):
     assert ps.ny == ss.ny - 1
     check_formulation(ps)
 
+    ps2 = PartiallySolvedSystem(ss, lambda x0, y0, p0, be: {
+        ss.dep[0]: y0[0]*be.exp(-p0[0]*(ss.indep-x0))})
+    assert ps2.ny == ss.ny - 1
+    check_formulation(ps2)
+
 
 @requires('sym', 'pyodeint')
 def test_PartiallySolvedSystem__by_name__from_linear_invariants():
@@ -1134,7 +1299,7 @@ def test_SymbolicSys__indep_in_exprs():
         return [t*p[0]*y[0]]
     be = sym.Backend('sympy')
     t, y, p = map(be.Symbol, 't y p'.split())
-    odesys = SymbolicSys([(y, dydt(t, [y], [p])[0])], t)
+    odesys = SymbolicSys([(y, dydt(t, [y], [p])[0])], t, params=True)
     fout = odesys.f_cb(2, [3], [4])
     assert len(fout) == 1
     assert abs(fout[0] - 2*3*4) < 1e-14
@@ -1143,6 +1308,7 @@ def test_SymbolicSys__indep_in_exprs():
 @requires('sym', 'pycvodes')
 @pytest.mark.parametrize('idx', [0, 1, 2])
 def test_PartiallySolvedSystem__roots(idx):
+    import sympy as sp
     t, x, y, z, p, q = sp.symbols('t x y z, p, q')
     odesys = SymbolicSys({x: -p*x, y: p*x - q*y, z: q*y}, t, params=(p, q), roots=([x - y], [x - z], [y - z])[idx])
     _p, _q, tend = 7, 3, 0.7
@@ -1160,6 +1326,7 @@ def test_PartiallySolvedSystem__roots(idx):
     check(psys)
 
 
+@pytest.mark.slow
 @requires('sym', 'pycvodes')
 @pytest.mark.parametrize('idx1,idx2,scaled,b2', product([0, 1, 2], [0, 1, 2], [True, False], [None, 0]))
 def test_TransformedSys__roots(idx1, idx2, scaled, b2):
@@ -1202,3 +1369,155 @@ def test_TransformedSys__roots(idx1, idx2, scaled, b2):
     })
     ploglog2 = LogLogSys.from_other(psys2)
     check(ploglog2)
+
+
+@requires('sym', 'sympy')
+def test__group_invariants():
+    be = sym.Backend('sympy')
+    x, y, z = symbs = be.symbols('x y z')
+
+    coeff1 = [3, 2, -1]
+    expr1 = 3*x + 2*y - z
+    lin, nonlin = _group_invariants([expr1], symbs, be)
+    assert lin == [coeff1]
+    assert nonlin == []
+
+    expr2 = 3*x*x + 2*y - z
+    lin, nonlin = _group_invariants([expr2], symbs, be)
+    assert lin == []
+    assert nonlin == [expr2]
+
+    lin, nonlin = _group_invariants([expr1, expr2], symbs, be)
+    assert lin == [coeff1]
+    assert nonlin == [expr2]
+
+    lin, nonlin = _group_invariants([x + be.exp(y)], symbs, be)
+    assert lin == []
+    assert nonlin == [x+be.exp(y)]
+
+
+@requires('sym', 'pycvodes')
+def test_SymbolicSys_as_autonomous():
+    import sympy
+
+    def rhs(t, y, p, backend=math):
+        return [y[1], backend.sin(t)-p[0]*y[0]]
+    odesys = SymbolicSys.from_callback(rhs, 2, 1)
+
+    def analytic(tout, init_y, p):
+        t, (k,) = odesys.indep, odesys.params
+        c1, c2 = sympy.symbols('c1 c2')
+        sqk = sympy.sqrt(k)
+        f = c1*sympy.cos(sqk*t) + c2*sympy.sin(sqk*t) + sympy.sin(t)/(k-1)
+        dfdt = f.diff(t)
+        t0 = tout[0]
+        sol, = sympy.solve([f.subs(t, t0) - init_y[0],
+                           dfdt.subs(t, t0) - init_y[1]],
+                           [c1, c2], dict=True)
+        sol[k] = p[0]
+        exprs = [f.subs(sol), dfdt.subs(sol)]
+        cb = sympy.lambdify([t], exprs)
+        return np.array(cb(tout)).T
+
+    def integrate_and_check(system):
+        init_y = [0, 0]
+        p = [2]
+        result = system.integrate([0, 80], init_y, p, integrator='cvode', nsteps=5000)
+        yref = analytic(result.xout, init_y, p)
+        assert np.all(result.yout - yref < 1.6e-5)
+
+    integrate_and_check(odesys)
+    assert len(odesys.dep) == 2
+    assert not odesys.autonomous_interface
+    assert not odesys.autonomous_exprs
+    asys = odesys.as_autonomous()
+    integrate_and_check(asys)
+    assert len(asys.dep) == 3
+    assert not asys.autonomous_interface
+    assert asys.autonomous_exprs
+
+
+@requires('sym', 'pycvodes')
+def test_SymbolicSys_as_autonomous__linear_invariants():
+    def rhs(t, y, p):
+        k = t**p[0]
+        return [-k*y[0], k*y[0]]
+
+    def analytic(tout, init_y, params):
+        y0ref = init_y[0]*np.exp(-tout**(params[0]+1)/(params[0]+1))
+        return np.array([y0ref, init_y[0] - y0ref + init_y[1]]).T
+
+    odes = SymbolicSys.from_callback(rhs, 2, 1, linear_invariants=[[1, 1]])
+    for odesys in [odes, odes.as_autonomous()]:
+        result = odesys.integrate(4, [5, 2], [3], integrator='cvode')
+        ref = analytic(result.xout, result.yout[0, :], result.params)
+        assert np.allclose(result.yout, ref, atol=1e-6)
+
+        invar_viol = result.calc_invariant_violations()
+        assert np.allclose(invar_viol, 0)
+
+
+@requires('sym', 'pycvodes')
+def test_SymbolicSys__by_name__as_autonomous():
+    def f(t, y, p):
+        k = t**p['e']
+        return {
+            'a': -k*y['a'],
+            'b': +k*y['a']
+        }
+
+    def analytic(tout, init_y, p):
+        y0ref = init_y[0]*np.exp(-tout**(p[0]+1)/(p[0]+1))
+        return np.array([y0ref, init_y[0] - y0ref + init_y[1]]).T
+
+    odes = SymbolicSys.from_callback(
+        f, names='ab', param_names='e', dep_by_name=True, par_by_name=True,
+        linear_invariants=[{'a': 1, 'b': 1}]
+    )
+
+    for odesys in [odes, odes.as_autonomous()]:
+        result = odesys.integrate(3, {'a': 2, 'b': 1}, {'e': 2}, integrator='cvode')
+        ref = analytic(result.xout, result.yout[0, :], result.params)
+        assert np.allclose(result.yout, ref, atol=1e-6)
+
+        invar_viol = result.calc_invariant_violations()
+        assert np.allclose(invar_viol, 0)
+
+
+@requires('sym', 'pycvodes')
+def test_SymbolicSys_as_autonomous__scaling():
+
+    # 2 HNO2 -> H2O + NO + NO2; MassAction(EyringHS.fk('dH1', 'dS1'))
+    # 2 NO2 -> N2O4; MassAction(EyringHS.fk('dH2', 'dS2'))
+    #
+    # HNO2 H2O NO NO2 N2O4
+    def get_odesys(scaling=1):
+        def rhs(t, y, p, backend=math):
+            HNO2, H2O, NO, NO2, N2O4 = y
+            dH1, dS1, dH2, dS2 = p
+            R = 8.314
+            T = 300 + 10*backend.sin(0.2*math.pi*t - math.pi/2)
+            kB_h = 2.08366e10
+            k1 = kB_h*T*backend.exp(dS1/R - dH1/(R*T))/scaling  # bimolecular => scaling**-1
+            k2 = kB_h*T*backend.exp(dS2/R - dH2/(R*T))/scaling  # bimolecular => scaling**-1
+            r1 = k1*HNO2**2
+            r2 = k2*NO2**2
+            return [-2*r1, r1, r1, r1 - 2*r2, r2]
+
+        return SymbolicSys.from_callback(rhs, 5, 4, names='HNO2 H2O NO NO2 N2O4'.split(),
+                                         param_names='dH1 dS1 dH2 dS2'.split())
+
+    def check(system, scaling=1):
+        init_y = [1*scaling, 55*scaling, 0, 0, 0]
+        p = [85e3, 10, 70e3, 20]
+        return system.integrate(np.linspace(0, 60, 200), init_y, p, integrator='cvode', nsteps=5000)
+
+    def compare_autonomous(scaling):
+        odesys = get_odesys(scaling)
+        autsys = odesys.as_autonomous()
+        res1 = check(odesys, scaling=scaling)
+        res2 = check(autsys, scaling=scaling)
+        assert np.allclose(res1.yout, res2.yout, atol=1e-6)
+
+    compare_autonomous(1)
+    compare_autonomous(1000)
